/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.combine;

import java.util.Collection;

/**
 * Abstract stub implementation of {@link CombineStrategy} that is intended to
 * be a convenient super-class for most implementations.
 *
 * @author Henrik Kaipe
 */
public abstract class AbstractCombineStrategy implements CombineStrategy {

    protected int maxCount = Integer.MAX_VALUE;
    
    private long randomSeed;
    private boolean randomSeedIsSet = false;

    public void setMaxCount(int maxCount) {
        this.maxCount = maxCount;
    }

    public void setRandomSeed(long randomSeed) {
        this.randomSeed = randomSeed;
        this.randomSeedIsSet = true;
    }
    
    public Collection combine(ValuesCollection vc) {
        final Collection records = combineInternal(vc);
        final int size = records.size();
        if (maxCount < size || Integer.MAX_VALUE == size) {
            throw new AssertionError("Number of callback records ("
                    + records.size() + ") exceeds the maximum number of "
                    + "callback records allowed (" + maxCount + "). If you "
                    + "think the maximum number of callback records is too low "
                    + "you can change it by reconfiguring the 'maxCount' "
                    + "parameter");
        } else {
            return records;
        }
    }

    /**
     * Returns the random-seed that has been set by
     * {@link #setRandomSeed(long)}.
     * If random-seed has not yet been set it will be set to the current
     * timestamp.
     */
    protected long getRandomSeed() {
        if (false == randomSeedIsSet) {
            setRandomSeed(System.currentTimeMillis());
            
            /*
             * By printing the generated random seed it will be documented and
             * could be reused in case the generated collection of
             * callback-records turns out the be of particular interest. (E.g.
             * by unexpectedly failing one of the tests.)
             */
            System.out.println("Combine using random seed: " + randomSeed);
        }
        return randomSeed;
    }

    /**
     * This is the method to implement when implementing
     * {@link CombineStrategy} by subclassing {@link AbstractCombineStrategy}.
     * The advantage compared to overriding {@link #combine(ValuesCollection)}
     * is that maxCount-check and validation of configuration for
     * {@link #setMaxCount(int)} and {@link #setRandomSeed(long)} will be taken
     * care of by the super-class - relieving this method to administer
     * combining exclusivly.
     */
    protected abstract Collection combineInternal(ValuesCollection vc);
}
