/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.combine;

import java.util.Collection;

/**
 * This is the API-interface that must be implemented by any
 * combine-strategy class.
 * <br/><br/>
 * The combine-strategy class should also have a default constructor or it
 * cannot be instanciated by the framework's built-in funtionality, which is
 * available through the class {@link
 * org.callbackparams.combine.annotation.AnnotationAwareCallbackRecordsFactory
 * AnnotationAwareCallbackRecordsFactory}, which assumes the annotation
 * {@link org.callbackparams.combine.annotation.CombineConfig @CombineConfig}
 * has been used to specify the combine-strategy class.
 * <br/><br/>
 * Instead of having the strategy-classes explicitly implement this interface
 * strategy-developers are encouraged to subclass
 * {@link AbstractCombineStrategy}. This abstract class offers some common
 * functionality and can also keep your test-classes up-to-date with any future
 * modifications to this interface.
 *
 * @see AbstractCombineStrategy
 *
 * @author Henrik Kaipe
 */
public interface CombineStrategy {

    /**
     * Specifies the maximum number of combinations that the method
     * {@link #combine(ValuesCollection)} is allowed to return. The default
     * value is implementation dependent.
     */
    void setMaxCount(int maxCount);
    
    /**
     * Some CombineStrategy implementations does to some extent pick
     * callback-record parameters randomly. With this setter it is possible to
     * specify which random seed to use and therewith make sure the exact same
     * set of callback-records is used on each test-run.
     */
    void setRandomSeed(long randomSeed);
    
    /**
     * Implementations of this method will create a collection of callback-
     * records based on the configuration of this CombineStrategy instance.
     * The callback-parameters used at a certain index in the created
     * callback-records are picked from the array at the corresponding index
     * in the argument ValuesCollection instance.
     */
    Collection combine(ValuesCollection vc);
}
