/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.internal;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.callbackparams.internal.template.TestrunCallbacks;
import org.callbackparams.support.MethodHashKey;

/**
 * @author Henrik Kaipe
 */
public abstract class CallbackContextBridge implements CallbackContext {

    private static final Constructor java5Constructor = lookupJava5Constr();
    private static Constructor lookupJava5Constr() {
        String className = CallbackContextBridge.class.getName();
        int endOfPackageName = className.lastIndexOf('.');
        String java5ClassName = className.substring(0, endOfPackageName)
                + ".annotation" + className.substring(endOfPackageName);
        try {
            final Constructor constr = Class.forName(java5ClassName)
                    .getDeclaredConstructor(new Class[] {
                ClassLoader.class, CallbackContext.class
            });
            constr.setAccessible(true);
            return constr;

        } catch (UnsupportedClassVersionError detectionOf_JDK1_4) {
//            System.out.println("Use CallbackContextBridge impl for JDK-1.4");
            return null;

        } catch (Exception shouldNeverHappen) {
            throw new Error(shouldNeverHappen);
        }
    }

    static CallbackContext newInstance(
            ClassLoader bridgedClassLoader, CallbackContext bridgedContext) {

        if (null != java5Constructor) {
            try {
                return (CallbackContext)
                        java5Constructor.newInstance(new Object[] {
                    bridgedClassLoader, bridgedContext
                });
            } catch (Exception shouldNeverHappen) {
                throw new Error(shouldNeverHappen);
            }

        } else {
            return new CallbackContextBridge(
                    bridgedClassLoader, bridgedContext) {};
        }
    }

    protected final CallbackContext bridgedContext;
    protected final ClassLoader bridgedLoader;
    protected final ClassLoader localLoader =
            TestrunCallbacks.class.getClassLoader();
    private final Map methodMaps = new HashMap();

    protected CallbackContextBridge(
            ClassLoader bridgedLoader,
            CallbackContext bridgedContext) {
        this.bridgedLoader = bridgedLoader;
        this.bridgedContext = bridgedContext;
    }

    public void addFieldToInject(Field f) {
        try {
            bridgedContext.addFieldToInject(
                    getBridgedDeclaringClass(f).getDeclaredField(f.getName()));
        } catch (NoSuchFieldException ex) {
            throw new Error(ex);
        }
    }

    public Iterator iterateFields() {
        return new BridgedMemberIterator(bridgedContext.iterateFields());
    }

    public int addMethodParams(Method m) {
        final Class bridgedClass = getBridgedDeclaringClass(m);
        return bridgedContext.addMethodParams( (Method)
                getMethodMap(bridgedClass).get(MethodHashKey.getHashKey(m)));
    }

    public Iterator iterateMethods() {
        return new BridgedMemberIterator(bridgedContext.iterateMethods());
    }

    public List getCallbackRecord() {
        return bridgedContext.getCallbackRecord();
    }

    public void setCallbackRecord(Collection callbackRecord) {
        List records = new ArrayList(callbackRecord.size());
        records.addAll(callbackRecord);
        bridgedContext.setCallbackRecord(records);
    }

    private Map getMethodMap(Class clazz) {
        if (false == methodMaps.containsKey(clazz)) {
            Map map = new HashMap();
            for (Iterator i = Arrays.asList(clazz.getDeclaredMethods())
                    .iterator() ; i.hasNext() ;) {
                final Method m = (Method) i.next();
                map.put(MethodHashKey.getHashKey(m), m);
            }
            methodMaps.put(clazz, map);
        }
        return (Map) methodMaps.get(clazz);
    }

    private Class getBridgedDeclaringClass(Member localMember) {
        try {
            return bridgedLoader
                    .loadClass(localMember.getDeclaringClass().getName());
        } catch (ClassNotFoundException ex) {
            throw new Error(ex);
        }
    }

    private class BridgedMemberIterator implements Iterator {
        private Iterator sourceMembers;

        public BridgedMemberIterator(Iterator sourceMembers) {
            this.sourceMembers = sourceMembers;
        }

        public boolean hasNext() {
            return sourceMembers.hasNext();
        }

        public Object next() {
            Member sourceMember = (Member) sourceMembers.next();
            Class localClass;
            try {
                localClass = localLoader.loadClass(
                        sourceMember.getDeclaringClass().getName());
            } catch (ClassNotFoundException ex) {
                throw new Error(ex);
            }

            if (sourceMember instanceof Field) {
                try {
                    return localClass.getDeclaredField(sourceMember.getName());
                } catch (NoSuchFieldException ex) {
                    throw new Error(ex);
                }

            } else if (sourceMember instanceof Method) {
                return getMethodMap(localClass)
                        .get(MethodHashKey.getHashKey((Method)sourceMember));

            } else {
                throw new UnsupportedOperationException(
                        "Unable to bridge members of "
                        + sourceMember.getClass());
            }
        }

        public void remove() {
            throw new UnsupportedOperationException("Not supported yet.");
        }
    }
}
