/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.internal.template;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.callbackparams.CallbackControlPanel;

/**
 * Wraps (by copying) an array of callback-parameters that are to be used for a
 * single test-run. This array is also referred to as a callback-record through-
 * out the callback-params documentation - hence the name of this class.
 *
 * @author Henrik Kaipe
 */
public class CallbackRecord {

    /**
     * The internally used list that contains instances of
     * {@link SingleParamCallbacks}.
     */
    private final List callbackRecord;
    private final Map latestCallbackResults = new IdentityHashMap();
    private final CallbackControlPanel ccp = new CallbackControlPanel() {
        public List getCurrentCallbackRecord() {
            return callbackRecord;
        }
        public Map getLatestCallbackResults() {
            return latestCallbackResults;
        }
    };
    private final String recordDescription;

    public CallbackRecord(Collection initialCallbackRecord) {
        this.callbackRecord = new ArrayList(initialCallbackRecord.size() + 5);
        this.callbackRecord.addAll(initialCallbackRecord);
        this.recordDescription = this.callbackRecord.toString();
    }

    public CallbackControlPanel getCallbackControlPanel() {
        return ccp;
    }

    public InvocationHandler newCallbackInvoker(final CallbackRef callbackRef) {
        return new InvocationHandler() {
            
            public Object invoke(Object o, Method m, Object[] args)
            throws Throwable {
                if (false == m.isAccessible()) {
                    m.setAccessible(true);
                }
                try {
                    latestCallbackResults.clear();

                    for (Iterator i = callbackRecord.iterator(); i.hasNext();) {
                        final Object callbackParameter = i.next();
                        if (null == callbackParameter) {
                            continue;
                        }
                        
                        final Object targetCallbackInstance =
                                callbackRef.asCallback(callbackParameter);
                        if (null != targetCallbackInstance) {
                            latestCallbackResults.put(
                                    callbackParameter,
                                    m.invoke(targetCallbackInstance, args));
                        }
                    }
                    return DefaultValue.forType(m.getReturnType());
                    
                } catch (final InvocationTargetException ite) {
                    throw ite.getTargetException();
                }
            }
        };
    }

    static class DefaultValue {
        private static final Map map = new HashMap();
        static {
            map.put(int.class, new Integer(0));
            map.put(boolean.class, Boolean.FALSE);
            map.put(long.class, new Long(0));
            map.put(char.class, new Character((char)0));
            map.put(double.class, new Double(0.0));
            map.put(byte.class, new Byte((byte)0));
            map.put(float.class, new Float(0.0F));
            map.put(short.class, new Short((short)0));
        }

        static Object forType(final Class type) {
            return map.get(type);
        } 
    }

    public String toString() {
        return recordDescription;
    }
}
