/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.internal.template;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Iterator;
import org.callbackparams.internal.CallbackContext;
import org.callbackparams.internal.CallbackContextFactory;
import org.callbackparams.internal.CallbackTestClassReloader;

/**
 * A fairly limited workaround implementation that can hopefully be improved as
 * time goes by. Its initial functionality was implemented so that integration
 * with the class-reloading mock-framework "Powermock" would be as seamless as
 * possible but if other class-reloading test-frameworks cause problems the
 * implementation should be improved to handle those cases as well.<br/><br/>
 * Unfortunatelly there is a major limitiation that hopefully will not turn out
 * to be that much of a problem. - Callback param-values other than
 * enum-constants are not supported ... It is one thing that enums are the
 * elegant way to construct param-values within code but if you wish to generate
 * your param-values from some non-java file (and have the values reached
 * through some static "values()" method) you have a problem. Hopefully the
 * future will bring us a better workaround ...
 *
 * @author Henrik Kaipe
 */
class ThirdPartyReloadingWorkaround {

    private static final String callbackParamsClassReloaderClassName =
            CallbackTestClassReloader.class.getName();

    static ThirdPartyReloadingWorkaround getInstance() {
        return new ThirdPartyReloadingWorkaround();
    }

    /**
     * The key to finding the callback-setup for this test-run is to find the
     * {@link CallbackTestClassReloader} class-loader in which the callback-
     * methods and callback-values were specified and then translate these
     * to their equivalents in the class-loader of the argument
     * "reloadedTestClass".<br/><br/>
     * The algorithm for finding the class-loader is
     * currently tailored in order to work in case the test-class was reloaded
     * by Powermock. This will likely work for a few other situations as well
     * but here might be a need for future improvements in case there comes a
     * situation that is much different from the Powermock case.<br/><br/>
     * As for translating the setup there are two kinds of objects that need
     * to be translated. The easy part is to translate the callback-methods
     * to their equivalent of the new class-loader - just look for the declaring
     * classes with the same name and identify their method with the same method
     * name and the same method-signature ... Translating the values in the
     * callback-record is a very different thing and the current solution only
     * manages to translate enum-values to their equivalent in the new
     * class-loader.
     */
    void resurrectSetup(Class reloadedTestClass)
    throws Exception {
        CallbackContext testRunContext = CallbackContextFactory
                .forClassLoader(reloadedTestClass.getClassLoader());

        ClassLoader callbackDefLoader = reloadedTestClass.getClassLoader();
        do {
            callbackDefLoader = callbackDefLoader.getClass().getClassLoader();
        } while (false == callbackDefLoader.getClass().getName()
                .equals(callbackParamsClassReloaderClassName));
        CallbackContext callbackDefContext = CallbackContextFactory
                .forClassLoader(callbackDefLoader);
        for (Iterator i = callbackDefContext.iterateMethods() ; i.hasNext() ;) {
            final Method m = (Method) i.next();
            testRunContext.addMethodParams(m);
        }
        for (Iterator i = callbackDefContext.iterateFields() ; i.hasNext() ;) {
            final Field f = (Field) i.next();
            testRunContext.addFieldToInject(f);
        }
        testRunContext.setCallbackRecord(
                callbackDefContext.getCallbackRecord());
    }
}
