/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.sandbox;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.beans.PropertyEditor;
import java.beans.PropertyEditorManager;
import java.beans.PropertyEditorSupport;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Proxy;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.lang.reflect.WildcardType;
import java.util.HashMap;
import java.util.Map;
import org.callbackparams.annotation.util.EnumConstantDescriptor;
import org.callbackparams.factory.CallbackFactory;

/**
 * This class has nested resources that can be used to produce callback-values
 * for {@link Property.Callback} through annotations. (The interface
 * {@link Property.Callback} itself cannot be implemented directly as of yet.)
 * <br/><br/>
 * Example usage ...
 * <pre>
 * &#64;RunWith({@link org.callbackparams.junit4.CallbackParamsRunner})
 * public class MyTest {
 *
 *   &#64;Test
 *   public void runTest(Property.Callback&lt;MyBean&gt; beanCallback) {
 *     // ...
 *   }
 *
 *   <b>&#64;Property.BeanClass(MyBean.class)
 *   &#64;Property.Name("myIntProperty")</b>
 *   enum MyEnum {
 *       <b>&#64;Property.Value("21")</b> FOO,
 *       <b>&#64;Property.Value("67")</b> BAR;
 *   }
 * }
 * </pre>
 * The annotations around and within <code>MyEnum</code> is meant as a shortcut
 * that is equivalent with this ...
 * <pre>
 *   enum MyEnum <b>implements Property.Callback&lt;MyBean&gt;</b> {
 *     FOO<b>(21)</b>, BAR<b>(67)</b>;
 *<b>
 *     int value;
 *
 *     MyEnum(int value) {
 *       this.value = value;
 *     }
 *
 *     void setValues(MyBean bean) {
 *         bean.setMyIntProperty(value);
 *     }
 *
 *     void assertValues(MyBean bean) {
 *       Assert.assertEquals("Value for property \"myIntProperty\" on " + bean,
 *           value, bean.getMyIntProperty());
 *     }</b>
 *   }
 * </pre>
 * Thus, a considerable amount of (<b>bold</b>) boilerplate code
 * can be excluded by using the
 * annotations, as shown in the first case.
 * <br/><br/>
 * This class is meant as a proof-of-concept of a backup-factory API that can
 * be used by a callback-interface in case a callback-value neither implements
 * the interface nor provides any implementation
 * through {@link org.callbackparams.CallbackFactory#getCallback()}.
 * <br/>
 * This class being a proof-of-concept here means that it applies to a
 * concept that is generally understood among developers (i.e. beans and
 * their properties) but unlike {@link Collector} the reusability of this
 * class is probably more limited.
 *
 * @author Henrik Kaipe
 * @see Collector
 */
public class Property<T> {

    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.TYPE)
    public @interface BeanClass {
        java.lang.Class value();
    }

    @Retention(RetentionPolicy.RUNTIME)
    @Target({ElementType.FIELD, ElementType.TYPE})
    public @interface Name {
        String value();
    }

    @Retention(RetentionPolicy.RUNTIME)
    @Target({ElementType.FIELD, ElementType.TYPE})
    public @interface Value {
        String value();
    }

    public interface Callback<T> {
        void setValues(T bean);
        void assertValues(T bean);

        /**
         * As of today the end-users are not supposed to have their callback-
         * values implement this interface and this method effectivly prevents
         * them from doing so by using API that cannot be accessed by end-user
         * code.
         * Unfortunatelly the method can be invoked, however, but that will
         * always generate an exception.
         * @param d parameter of a class with private access, which cannot
         *          be reached by end-user code
         * @throws UnsupportedOperationException always
         * @deprecated we will hopefully find a better solution for this pretty
         *             ugly workaround
         */
        void dummyMethodThatMakesThisInterfaceImpossibleToImplement(
                PropertyCallbackFactory.Dummy d)
        throws UnsupportedOperationException;

        /**
         * It is because of this inner class that Property.Callback parameters
         * will have their callbacks constructed in a very different way.
         */
        class PropertyCallbackFactory implements CallbackFactory {

            final Class<?> beanClass;
            final Class[] parameterInterfaces;
            final Map<String,PropertyDescriptor> propertyDescriptors =
                    new HashMap<String, PropertyDescriptor>();

            private Type eraseAllButTypeVariable(Type type) {

                if (type instanceof WildcardType) {
                    /* Check for a lower bound */
                    Type[] lowerBounds = ((WildcardType)type).getLowerBounds();
                    if (0 == lowerBounds.length) {
                        return null;
                    } else {
                        type = lowerBounds[0];
                    }
                }

                if (type instanceof ParameterizedType) {
                    return ((ParameterizedType)type).getRawType();

                } else {
                    /* Class of type-variable ... */
                    return type;
                }
            }

            /**
             * Recursive thing that is supposed to make us find the
             */
            private Type determineGenericTypeDeclaration(Type type) {
                if (Callback.class == type) {
                    throw new IllegalArgumentException(
                            "The Property.Callback parameter declaration must "
                            + "specify the bean-class as the generic type-"
                            + "parameter of the method paramter-declarations."
                            + " E.g. 'testMethod(Property.Callback<BeanClass> "
                            + "propertyCallback)");

                }

                Class<?> rawType = (Class<?>) (type instanceof Class
                        ? type : ((ParameterizedType) type).getRawType());

                if (Callback.class == rawType) {
                    return eraseAllButTypeVariable( ((ParameterizedType)type)
                            .getActualTypeArguments()[0]);

                } else if (Callback.class.isAssignableFrom(rawType)) {
                    for (Type interf : rawType.getGenericInterfaces()) {
                        final Type beanType =
                                determineGenericTypeDeclaration(interf);

                        if (null == beanType) {
                            continue;

                        } else if (beanType instanceof Class) {
                            return beanType;

                        } else if (rawType == ((TypeVariable<?>)beanType)
                                .getGenericDeclaration()) {
                            final TypeVariable<?>[] typeParameters =
                                    rawType.getTypeParameters();
                            for (int i = 0 ; i < typeParameters.length ; ++i) {
                                if (beanType == typeParameters[i]) {
                                    return eraseAllButTypeVariable(
                                            ((ParameterizedType)type)
                                            .getActualTypeArguments()[i]);
                                }
                            }
                            throw new Error(beanType + " does not seem to be " +
                                    "declared by " + rawType);

                        } else {
                            return beanType;
                        }
                    }
                    return null;

                } else {
                    return null;
                }
            }

            public PropertyCallbackFactory(Method m, int paramIndex) {
                Type beanType =  determineGenericTypeDeclaration(
                        m.getGenericParameterTypes()[paramIndex]);
                if (null == beanType) {
                    throw new Error("Could not determine specified bean-class"
                            + " for parameter-declaration "
                            + m.getGenericParameterTypes()[paramIndex]);
                } else {
                    System.err.println("Bean type: " + beanType);
                    beanClass = (Class<?>) beanType;
                    parameterInterfaces =
                            new Class[] {m.getParameterTypes()[paramIndex]};
                    try {
                        for (PropertyDescriptor pd :
                                Introspector.getBeanInfo(beanClass)
                                .getPropertyDescriptors()) {
                            if (null == pd.getPropertyEditorClass()) {
                                Object tmpEditor = PropertyEditorManager
                                        .findEditor(pd.getPropertyType());
                                pd.setPropertyEditorClass(null != tmpEditor
                                        ? tmpEditor.getClass()
                                        : PropertyEditorSupport.class);
                            }
                            propertyDescriptors.put(pd.getName(), pd);
                        }
                    } catch (IntrospectionException ex) {
                        throw new Error(ex);
                    }
                }
            }

            public Object asCallback(Object callbackRecordElement) {
                if (false == callbackRecordElement instanceof Enum) {
                    return null;
                }

                EnumConstantDescriptor ecd = new EnumConstantDescriptor(
                        (Enum<?>) callbackRecordElement);

                Value value = ecd.getAnnotation(Value.class);
                if (null == value) {
                    return null;
                }
                BeanClass beanClass = ecd.getAnnotation(BeanClass.class);
                if (null == beanClass || this.beanClass != beanClass.value()) {
                    return null;
                }

                PropertyDescriptor pd = propertyDescriptors.get(
                        ecd.getAnnotation(Name.class, true).value());
                final BasicImpl impl = new BasicImpl(pd, value.value());

                if (Callback.class == parameterInterfaces[0]) {
                    return impl;

                } else {
                    return Proxy.newProxyInstance(
                            parameterInterfaces[0].getClassLoader(),
                            parameterInterfaces,
                            new InvocationHandler() {
                        public Object invoke(
                                Object proxy, Method method, Object[] args)
                        throws Throwable {
                            try {
                                return method.invoke(impl, args);
                            } catch (InvocationTargetException x) {
                                throw x.getTargetException();
                            }
                        }
                    });
                }
            }

            private static class Dummy {}
        }
    }

    private static class BasicImpl extends Impl {

        final PropertyDescriptor descriptor;
        final Object[] setterArgument;

        public BasicImpl(PropertyDescriptor descriptor, String value) {
            this.descriptor = descriptor;
            PropertyEditor editor = PropertyEditorManager
                    .findEditor(descriptor.getPropertyType());
            editor.setAsText(value);
            this.setterArgument = new Object[] {editor.getValue()};
        }

        Object invoke(Method m, Object bean, Object[] args) {
            try {
                try {
                    m.setAccessible(true);
                } catch (SecurityException ignore) {
                    /* Never mind */
                }
                return m.invoke(bean, args);

            } catch (Exception x) {
                Throwable toThrow;
                if (x instanceof InvocationTargetException) {
                    toThrow = ((InvocationTargetException)x)
                            .getTargetException();
                } else {
                    toThrow = x;
                }
                if (toThrow instanceof RuntimeException) {
                    throw (RuntimeException)toThrow;
                } else {
                    throw new Error(toThrow);
                }
            }
        }

        public void setValues(Object bean) {
            invoke(descriptor.getWriteMethod(), bean, setterArgument);
        }

        public void assertValues(Object bean) {
            Object actual = invoke(descriptor.getReadMethod(), bean, null);
            assert setterArgument[0].equals(actual) : "Value for property \""
                    + descriptor.getName() + "\" on " + bean + ": expected <"
                    + setterArgument[0] + "> but was <" + actual + ">";
        }
    }

    public abstract static class Impl implements Callback<Object> {

        /**
         * @deprecated we will hopefully find a better solution for this pretty
         *             ugly workaround
         */
        public final void dummyMethodThatMakesThisInterfaceImpossibleToImplement(
                PropertyCallbackFactory.Dummy d) {
            throw new UnsupportedOperationException(
                    "This method is not supposed to be used");
        }
    }

    private Property() {}
}
